package ru.dbotthepony.mc.otm.config

import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.defineDecimal
import ru.dbotthepony.mc.otm.registry.MNames

object ItemsConfig : AbstractConfig("items") {
	private fun verboseValues(name: String, storage: Decimal, receive: Decimal, extract: Decimal = receive): VerboseBalanceValues {
		builder.push(name)

		val obj = object : VerboseBalanceValues {
			override val capacity: Decimal by builder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val receive: Decimal by builder.defineDecimal("receive", receive, minimum = Decimal.ONE)
			override val extract: Decimal by builder.defineDecimal("extract", extract, minimum = Decimal.ONE)
		}

		builder.pop()

		return obj
	}

	private fun batteryValues(name: String, storage: Decimal, receive: Decimal, extract: Decimal = receive, initialBatteryLevel: Decimal = Decimal.ZERO): BatteryBalanceValues {
		builder.push(name)

		val obj = object : BatteryBalanceValues {
			override val capacity: Decimal by builder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val receive: Decimal by builder.defineDecimal("receive", receive, minimum = Decimal.ONE)
			override val extract: Decimal by builder.defineDecimal("extract", extract, minimum = Decimal.ONE)
			override val initialBatteryLevel: Decimal by builder.defineDecimal("initialBatteryLevel", initialBatteryLevel, minimum = Decimal.ZERO)
		}

		builder.pop()

		return obj
	}

	private fun conciseValues(name: String, storage: Decimal, throughput: Decimal): ConciseBalanceValues {
		builder.push(name)

		val obj = object : ConciseBalanceValues {
			override val capacity: Decimal by builder.defineDecimal("capacity", storage, minimum = Decimal.ONE)
			override val throughput: Decimal by builder.defineDecimal("throughput", throughput, minimum = Decimal.ONE)
		}

		builder.pop()

		return obj
	}

	init {
		builder.push("EnergyBatteries")
	}

	object Batteries {
		val CRUDE = batteryValues(MNames.BATTERY_CRUDE, Decimal(100_000), Decimal(160), Decimal(40), Decimal(80_000))
		val BASIC = batteryValues(MNames.BATTERY_BASIC, Decimal(400_000), Decimal(600))
		val NORMAL = batteryValues(MNames.BATTERY_NORMAL, Decimal(2_000_000), Decimal(1_000))
		val DENSE = batteryValues(MNames.BATTERY_DENSE, Decimal(10_000_000), Decimal(2_000))
		val CAPACITOR = batteryValues(MNames.BATTERY_CAPACITOR, Decimal(500_000), Decimal(50_000))

		val QUANTUM_BATTERY = conciseValues(MNames.QUANTUM_BATTERY, Decimal(40_000_000), Decimal(10_000))
		val QUANTUM_CAPACITOR = conciseValues(MNames.QUANTUM_CAPACITOR, Decimal(1_000_000), Decimal(200_000))

		val ZPM = conciseValues(MNames.ZPM_BATTERY, Decimal(200_000_000_000_000L), Decimal(200_000_000L))
	}

	init {
		Batteries
		builder.pop()
	}

	val MATTER_DUST_CAPACITY by builder.comment("Maximal matter value one matter dust item can have").defineDecimal("matterDustCapacity", Decimal(2_000), minimum = Decimal.ONE_TENTH)

	init {
		builder.push("MatterCapacitors")
	}

	object Capacitors {
		val BASIC by builder.defineDecimal(MNames.MATTER_CAPACITOR_BASIC, Decimal(2_500), minimum = Decimal.ONE_TENTH)
		val NORMAL by builder.defineDecimal(MNames.MATTER_CAPACITOR_NORMAL, Decimal(10_000), minimum = Decimal.ONE_TENTH)
		val DENSE by builder.defineDecimal(MNames.MATTER_CAPACITOR_DENSE, Decimal(25_000), minimum = Decimal.ONE_TENTH)
	}

	init {
		Capacitors
		builder.pop()
	}

	object PatternDrives {
		val NORMAL: Int by builder.defineInRange(MNames.PATTERN_DRIVE_NORMAL, 4, 1, Int.MAX_VALUE)
	}

	init {
		PatternDrives
	}
}
