package ru.dbotthepony.mc.otm.core.util

import com.google.common.collect.ImmutableList
import ru.dbotthepony.mc.otm.core.math.Decimal
import java.math.BigDecimal
import java.math.BigInteger

enum class SiPrefix(
	val power: Int,
	fractional: Boolean,
	val symbol: Char,
) {
	// multiplies
	KILO (3,  false, 'k'),
	MEGA (6,  false, 'M'),
	GIGA (9,  false, 'G'),
	TERA (12, false, 'T'),
	PETA (15, false, 'P'),
	EXA  (18, false, 'E'),
	ZETTA(21, false, 'Z'),
	YOTTA(24, false, 'Y'),

	// decimals
	// DECI (1,  true,  'd'),
	// CENTI(2,  true,  'c'),
	MILLI(3,  true,  'm'),
	MICRO(6,  true,  'μ'),
	NANO (9,  true,  'n'),
	PICO (12, true,  'p'),
	FEMTO(15, true,  'f'),
	ATTO (18, true,  'a'),
	ZEPTO(21, true,  'z'),
	YOCTO(24, true,  'y');

	val formatLocaleKey = "otm.suffix.${name.lowercase()}".intern()
	val rawLocaleKey = "otm.suffix_raw.${name.lowercase()}".intern()

	val string = if (fractional) "0." + "0".repeat(power - 1) + "1" else "1" + "0".repeat(power)

	fun paddedIndex(input: String, index: Int): Char {
		val finalIndex = input.length - power + index

		if (finalIndex >= 0) {
			return input[finalIndex]
		}

		return '0'
	}

	val decimal = BigDecimal(string)
	val impreciseFraction = Decimal(string)
	val integer = if (!fractional) BigInteger(string) else null

	val long = if (!fractional) string.toLongOrNull() else null
	val int = if (!fractional) string.toIntOrNull() else null
	val double = string.toDouble()

	companion object {
		@JvmField
		val MULTIPLIES: List<SiPrefix> = ImmutableList.builder<SiPrefix>()
			.add(KILO)
			.add(MEGA)
			.add(GIGA)
			.add(TERA)
			.add(PETA)
			.add(EXA)
			.add(ZETTA)
			.add(YOTTA)
			.build()

		@JvmField
		val DECIMALS: List<SiPrefix> = ImmutableList.builder<SiPrefix>()
			//.add(DECI)
			//.add(CENTI)
			.add(MILLI)
			.add(MICRO)
			.add(NANO)
			.add(PICO)
			.add(FEMTO)
			.add(ATTO)
			.add(ZEPTO)
			.add(YOCTO)
			.build()

		@JvmField
		val DECIMALS_IMPRECISE: List<SiPrefix> = ImmutableList.builder<SiPrefix>()
			//.add(DECI)
			//.add(CENTI)
			.add(MILLI)
			.add(MICRO)
			.build()
	}
}

